// Copyright 2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package edu.berkeley.sbp.tib;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.chr.*;
import java.util.*;
import java.io.*;

// TODO: multiple {{ }} for superquotation
// TODO: strings
// TODO: comments

/**
 *   A slow, ugly, inefficient, inelegant, ad-hoc parser for TIB files.
 *
 *   Despite being inelegant, this parser keeps all internal state on
 *   the stack (the only heap objects created are those which are
 *   returned).
 *
 *   This was written as an ad-hoc parser to facilitate
 *   experimentation with the TIB spec.  Once the spec is finalized it
 *   should probably be rewritten.
 */
public class Tib implements Input<Character> {

    public Tib(String s) throws IOException { this(new StringReader(s)); }
    public Tib(Reader r) throws IOException { this(new BufferedReader(r)); }
    public Tib(InputStream is) throws IOException { this(new BufferedReader(new InputStreamReader(is))); }
    public Tib(BufferedReader br) throws IOException {
        this.br = br;
        istack.add(-1);
        //cur = parse(br);
        //System.out.println("\rparsing: \"" + cur.toString(0, -1) + "\"");
    }

    private String s = "";
    int pos = 0;
    int spos = 0;

    int _row = 1;
    int _col = 0;
    public Input.Location getLocation() { return new CartesianLocation(_row, _col); }
    private BufferedReader br;

    char left = CharRange.left;
    char right = CharRange.right;

    boolean waiting = false;
    char waitingChar = ' ';
    boolean indenting = true;
    int indentation = 0;
    private ArrayList<Integer> istack = new ArrayList<Integer>();
    public Character next(int numstates, int resets, int waits) throws IOException {
        Character ret = nextc(numstates, resets);
        if      (ret==left)  System.out.print("\033[31m{\033[0m");
        else if (ret==right) System.out.print("\033[31m}\033[0m");
        else if (ret==null) return null;
        else System.out.print(ret);
        return ret;
    }

    Character waitingBrace = null;
    public Character nextc(int numstates, int resets) throws IOException {
        char c;
        if (waitingBrace != null) {
            Character ret = waitingBrace;
            waitingBrace = null;
            return ret;
        }
        if (waiting) {
            waiting = false;
            c = waitingChar;
        } else {
            int i = br.read();
            if (i==-1) {
                if (istack.size() > 1) {
                    istack.remove(istack.size()-1);
                    return right;
                }
                return null;
            }
            c = (char)i;
        }
        if (c=='\n') { _row++; _col=0; }
        else         _col++;
        if (indenting) {
            if (c==' ') { indentation++; return done(c); }
            if (c=='\n') { indentation = 0; if (blank) return nextc(numstates, resets); blank = true; waiting = true; waitingChar='\n'; return '\n'; }
            int last = istack.size()==0 ? -1 : istack.get(istack.size()-1);
            if (indentation==last) {
                if (blank) {
                    indenting = false;
                    waitingChar = c;
                    waiting = true;
                    waitingBrace = left;
                    return right;
                    //return nextc(numstates);
                }
                blank = false;
                indenting = false;
                return done(c);
            }
            blank = false;
            waitingChar = c;
            waiting = true;
            if (indentation > last) {
                indenting = false;
                istack.add(indentation);
                //System.out.print("\033[31m+"+indentation+"+\033[0m");
                return left;
            } else /*if (indentation < last)*/ {
                istack.remove(istack.size()-1);
                //System.out.print("\033[31m-"+last+"-\033[0m");
                blank = true;
                return right;
            }
        } else {
            blank = false;
            if (c=='\n') { indenting=true; indentation = 0; }
            return done(c);
        }
    }

    public Character done(char c) {
        switch(c) {
            case '{': return left;
            case '}': return right;
            default: return c;
        }
    }
    boolean blank = false;

    // Grammar //////////////////////////////////////////////////////////////////////////////

    public static class Grammar extends MetaGrammar {
        private int anon = 0;
        private final Element ws = Repeat.maximal0(nonTerminal("w"));
        public Grammar() { dropAll.add(ws); }
        public Object walk(Tree<String> tree) {
            String head = tree.head();
            if (tree.numChildren()==0) return super.walk(tree);
            if ("{".equals(head)) {
                String s = "braced"+(anon++);
                Union u = nonTerminal(s);
                Union u2 = ((PreSequence)walk(tree, 0)).sparse(ws).buildUnion();
                u2.add(Sequence.singleton(new Element[] { u }, 0, null, null));
                return nonTerminal(s,
                                   new PreSequence[][] {
                                       new PreSequence[] {
                                           new PreSequence(new Element[] { CharRange.leftBrace,
                                                                           ws,
                                                                           u2,
                                                                           ws,
                                                                           CharRange.rightBrace
                                           })
                                       }
                                   },
                                   false,
                                   false);
            }
            return super.walk(tree);
        }
    }

    /*
public class Braces extends Union {

    private static final Element left  = string("{");
    private static final Element right = string("}");
    
    public static String join(Object[] e) {
        StringBuffer ret = new StringBuffer();
        for(int i=0; i<e.length; i++) {
            if (i>0) ret.append(" ");
            ret.append(e[i]);
        }
        return ret.toString();
    }

    public Braces(Element[] e, Element sep) {
        super("{"+join(e)+"}");
        Element[] e2 = new Element[sep == null ? e.length+2 : e.length + 4];
        e2[0]           = left;
        e2[e2.length-1] = right;
        if (sep != null) {
            e2[1] = sep;
            e2[e2.length-2] = sep;
        }
        for(int i=0; i<e.length; i++) e2[i+(sep==null ? 1 : 2)] = e[i];
        addAlternative(new Sequence.Singleton(e2));
        addAlternative(new Sequence.Singleton(sep==null ? new Element[] { left, this, right} : new Element[] { left, sep, this, sep, right }));
    }
}
                      */

}

